with Windows.Strings;
with System;

package body Windows.Processes is

   -- ------------------------------------------------------------------------

   subtype Environment_Block_Type   is Strings.Environment_Block_Type;
   subtype Environment_Block_Access is Strings.Environment_Block_Access;
   subtype API_String_Type          is Strings.API_String_Type;
   subtype API_String_Access        is Strings.API_String_Access;

   -- ------------------------------------------------------------------------

   function Create_Process
     (Application_Name : API_String_Access;
      Command_Line : API_String_Access;
      Process_Security_Attributes : access Security_Attributes_Type;
      Primary_Thread_Security_Attributes : access Security_Attributes_Type;
      Grant_Access_To_Inheritable_Objects : Boolean_Type;
      Creation_Mode : Creation_Mode_Type;
      Environment_Block : Environment_Block_Access;
      Startup_Directory : API_String_Access;
      Startup_Informations : System.Address;
      Process_Informations_Result : not null access Process_Informations_Type)
      return Boolean_Type;

   pragma Import
     (Convention    => Stdcall,
      Entity        => Create_Process,
      External_Name => "CreateProcessW");

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line                  : String;
      Startup_Directory             : String;
      No_Startup_Directory_If_Empty : Boolean;
      Startup_Informations          : Startup_Informations_Type;
      Startup_Options               : Startup_Options_Type)
      return Process_Informations_Type
   is

      Command_Line_API_String : constant Strings.API_String_Type :=
        Strings.Create
          (Source        => Command_Line,
           Null_If_Empty => False);

      Command_Line_Access : constant API_String_Access :=
        Strings.Access_To (Command_Line_API_String);

      -- ---------------------------------------------------------------------

      Process_Security_Attributes_Copy : aliased Security_Attributes_Type;

      function Process_Security_Attributes_Discriminated_Access
        return access Security_Attributes_Type
      is
      begin

         if Startup_Options.Process_Security_Attributes_Defined then
            Process_Security_Attributes_Copy :=
              (Startup_Options.Process_Security_Attributes);
            return Process_Security_Attributes_Copy'Access;
         else
            return null;
         end if;
      end;

      Process_Security_Attributes_Access : access Security_Attributes_Type :=
        (Process_Security_Attributes_Discriminated_Access);

      -- ---------------------------------------------------------------------

      Primary_Thread_Security_Attributes_Copy
        : aliased Security_Attributes_Type;

      function Primary_Thread_Security_Attributes_Discriminated_Access
        return access Security_Attributes_Type
      is
      begin

         if Startup_Options.Primary_Thread_Security_Attributes_Defined then
            Process_Security_Attributes_Copy :=
              (Startup_Options.Primary_Thread_Security_Attributes);
            return Primary_Thread_Security_Attributes_Copy'Access;
         else
            return null;
         end if;
      end;

      Primary_Thread_Security_Attributes_Access
        : access Security_Attributes_Type :=
          (Primary_Thread_Security_Attributes_Discriminated_Access);

      -- ---------------------------------------------------------------------

      Environment_Block_API_String : constant Environment_Block_Type :=
        Strings.Create
          (Source        => Startup_Options.Environment_Variables,
           Null_If_Empty => True);

      Environment_Block : constant Environment_Block_Access :=
        Strings.Access_To_Or_Null
          (Environment_Block => Environment_Block_API_String,
           Null_Condition    =>
             (not Startup_Options.Environment_Variables_Defined));

      -- ---------------------------------------------------------------------

      Startup_Directory_API_String : constant API_String_Type :=
        Strings.Create
          (Source        => Startup_Directory,
           Null_If_Empty => No_Startup_Directory_If_Empty);

      Startup_Directory_Access   : constant API_String_Access :=
        Strings.Access_To (Startup_Directory_API_String);

      use type System.Address;
   begin
      if Startup_Informations'Address = System.Null_Address then
         raise Error;
      end if;

      return Result : aliased Process_Informations_Type do
         declare
            Operation_Status : constant Boolean_Type :=
              Create_Process
                (Application_Name => Strings.Null_String,

                 Command_Line => Command_Line_Access,

                 Process_Security_Attributes =>
                   (Process_Security_Attributes_Access),

                 Primary_Thread_Security_Attributes =>
                   (Primary_Thread_Security_Attributes_Access),

                 Grant_Access_To_Inheritable_Objects =>
                   (From_Boolean
                      (Startup_Options.Grant_Access_To_Inheritable_Objects)),

                 Creation_Mode => Startup_Options.Creation_Mode,

                 Environment_Block => Environment_Block,

                 Startup_Directory => Startup_Directory_Access,

                 Startup_Informations => Startup_Informations'Address,

                 Process_Informations_Result => Result'Access);
         begin
            if Operation_Status = Boolean_False then
               raise Error;
            end if;
         end;
      end return;
   end;

   -- ------------------------------------------------------------------------

   Empty_Startup_Directory : constant String := "";

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Process_Informations_Type
   is
   begin
      return Run
        (Command_Line                  => Command_Line,
         Startup_Directory             => Empty_Startup_Directory,
         No_Startup_Directory_If_Empty => True,
         Startup_Informations          => Startup_Informations,
         Startup_Options               => Startup_Options);
   end;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Directory    : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Process_Informations_Type
   is
   begin
      return Run
        (Command_Line                  => Command_Line,
         Startup_Directory             => Startup_Directory,
         No_Startup_Directory_If_Empty => False,
         Startup_Informations          => Startup_Informations,
         Startup_Options               => Startup_Options);
   end;

   -- ------------------------------------------------------------------------

   type Wait_Result_Kinds is
     (Signaled,
      Abandoned,
      Timeout,
      Failed);

   for Wait_Result_Kinds use
     (
      Signaled  => 16#00000000#,
      Abandoned => 16#00000080#,
      Timeout   => 16#00000102#,
      Failed    => 16#FFFFFFFF#
     );

   for Wait_Result_Kinds'Size use Unsigned_Type'Size;

   -- ------------------------------------------------------------------------

   type Timeout_Delay_Type is new Unsigned_Type;
   Inifinite_Delay : constant Timeout_Delay_Type := 16#FFFFFFFF#;

   -- ------------------------------------------------------------------------

   function Wait
     (Process       : Process_Type;
      Timeout_Delay : Timeout_Delay_Type := Inifinite_Delay)
      return Wait_Result_Kinds
   is

      function Wait
        (Process       : Process_Type;
         Timeout_Delay : Timeout_Delay_Type := Inifinite_Delay)
      return Wait_Result_Kinds;

      pragma Import
        (Convention    => Stdcall,
         Entity        => Wait,
         External_Name => "WaitForSingleObject");

   begin
      return Result : constant Wait_Result_Kinds :=
        (Wait (Process => Process, Timeout_Delay => Timeout_Delay))
      do
         if not Result'Valid then
            raise Error;
         end if;
      end return;
   end;

   -- ------------------------------------------------------------------------

   function Get_Process_Exit_Code
     (Process   : in Process_Type;
      Exit_Code : not null access Exit_Code_Type)
      return Boolean_Type;

   pragma Import
     (Convention    => Stdcall,
      Entity        => Get_Process_Exit_Code,
      External_Name => "GetExitCodeProcess");

   function Process_Exit_Code (Process : Process_Type) return Exit_Code_Type
   is
   begin
      return Exit_Code : aliased Exit_Code_Type do
         declare
            Operation_Status : constant Boolean_Type :=
              Get_Process_Exit_Code
                (Process => Process,
                 Exit_Code => Exit_Code'Access);
         begin
            if Operation_Status = Boolean_False then
               raise Error;
            end if;
         end;
      end return;
   end;

   -- ------------------------------------------------------------------------

   function Close (Process : Process_Type) return Boolean_Type;

   pragma Import
     (Convention    => Stdcall,
      Entity        => Close,
      External_Name => "CloseHandle");

   procedure Close (Process : Process_Type)
   is
      Operation_Status : constant Boolean_Type := Close (Process);
   begin
      if Operation_Status = Boolean_False then
         raise Error;
      end if;
   end;

   -- ------------------------------------------------------------------------

   procedure Wait_Get_Exit_Code_And_Close
     (Process_Informations : in out Process_Informations_Type;
      Exit_Code            : out Exit_Code_Type)
   is
   begin

      declare
         Wait_Result : constant Wait_Result_Kinds :=
           (Wait (Process_Informations.Process));
      begin
         case Wait_Result is
            when Signaled =>
               null;
            when Abandoned =>
               raise Error;
            when Timeout =>
               raise Error;
            when Failed =>
               raise Error;
         end case;
      end;

      Exit_Code := Process_Exit_Code (Process_Informations.Process);

      Close (Process_Informations.Process);

   end;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Exit_Code_Type
   is
      Process_Informations : Process_Informations_Type :=
        Run
          (Command_Line         => Command_Line,
           Startup_Informations => Startup_Informations,
           Startup_Options      => Startup_Options);
   begin
      return Result : Exit_Code_Type do
         Wait_Get_Exit_Code_And_Close
           (Process_Informations => Process_Informations,
            Exit_Code            => Result);
      end return;
   end;

   -- ------------------------------------------------------------------------

   function Run
     (Command_Line         : String;
      Startup_Directory    : String;
      Startup_Informations : Startup_Informations_Type;
      Startup_Options      : Startup_Options_Type)
      return Exit_Code_Type
   is
      Process_Informations : Process_Informations_Type :=
        Run
          (Command_Line         => Command_Line,
           Startup_Directory    => Startup_Directory,
           Startup_Informations => Startup_Informations,
           Startup_Options      => Startup_Options);
   begin
      return Result : Exit_Code_Type do
         Wait_Get_Exit_Code_And_Close
           (Process_Informations => Process_Informations,
            Exit_Code            => Result);
      end return;
   end;

end Windows.Processes;
