package body Windows.Strings is

   Command_Line : API_String_Access := null;

   pragma Export
     (Convention    => C,
      Entity        => Command_Line,
      External_Name => "commandline");

   -- ------------------------------------------------------------------------

   type Allocation_Type is record
      Used : Boolean           := False;
      Data : API_String_Access := null;
   end record;

   -- ------------------------------------------------------------------------

   function Capacity (Allocation : Allocation_Type) return Data_Length_Type is
   begin
      if Allocation.Data'Length = 0 then
         raise Error;
      end if;

      return Allocation.Data'Length - 1;
   end Capacity;

   -- ------------------------------------------------------------------------

   Static_Allocations : array (1 .. 9) of Allocation_Type;

   -- ------------------------------------------------------------------------

   function Free_Static_Allocation_For
     (Length : Data_Length_Type)
      return   API_String_Access
   is
   begin
      for Index in Static_Allocations'Range loop
         if Static_Allocations (Index).Data /= null then
            if Capacity (Static_Allocations (Index)) >= Length and
               not Static_Allocations (Index).Used
            then
               Static_Allocations (Index).Used := True;
               return Static_Allocations (Index).Data;
            end if;
         end if;
      end loop;

      raise Error;
   end Free_Static_Allocation_For;

   -- ------------------------------------------------------------------------

   procedure Initialize (S : in out API_String_Type) is
   begin
      Finalization.Initialize (Finalization.Limited_Controlled (S));
      S.String_Data := null;
   end Initialize;

   -- ------------------------------------------------------------------------

   procedure Finalize (S : in out API_String_Type) is
   begin
      if S.String_Data = null then
         null;
      elsif S.String_Data = Command_Line then
         null;
      else
         Iterate_Allocation_Pool : for Index in Static_Allocations'Range loop
            if S.String_Data = Static_Allocations (Index).Data then
               Static_Allocations (Index).Used := False;

               S.String_Data := null;
               S.Data_Length := 0;
               exit Iterate_Allocation_Pool;
            end if;
         end loop Iterate_Allocation_Pool;

         if S.String_Data /= null then
            raise Error;
         end if;
      end if;

      Finalization.Finalize (Finalization.Limited_Controlled (S));
   end Finalize;

   -- ------------------------------------------------------------------------

   type Word24_Type is mod 2 ** 24;
   type Byte_Type is mod 256;

   Hight_Bit_Mask   : constant Byte_Type := 2#10000000#;
   Low_Bits_Mask    : constant Byte_Type := 2#00111111#;
   Hight_Bits_Mask  : constant Byte_Type := 2#11000000#;
   Hight_Bits_Value : constant Byte_Type := 2#10000000#;

   -- ------------------------------------------------------------------------

   function Length_In_UTF16 (Source : String) return Data_Length_Type is
      Size      : Data_Length_Type;
      Remaining : Natural;
      Byte      : Byte_Type;
   begin
      if Source'Length = 0 then
         return 0;
      end if;

      Size      := 0;
      Remaining := 0;

      for Index in Source'First .. Source'Last loop
         if Remaining > 0 then
            Remaining := Remaining - 1;
         else
            Byte := Character'Pos (Source (Index));
            if (Byte and Hight_Bit_Mask) = 0 then
               if Size = Data_Length_Type'Last then
                  raise Error;
               end if;

               Size      := Size + 1;
               Remaining := 0;
            else
               Byte := Byte * 2;
               Remaining := 0;

               while (Byte and Hight_Bit_Mask) /= 0 loop
                  Remaining := Remaining + 1;
                  Byte      := Byte * 2;
               end loop;

               if Byte = 0 then
                  raise Error;
               end if;

               if Remaining > 4 then
                  raise Error;
               elsif Remaining > 2 then
                  if Size >= Data_Length_Type'Last - 1 then
                     raise Error;
                  end if;
                  Size := Size + 2;
               elsif Remaining > 0 then
                  if Size = Data_Length_Type'Last then
                     raise Error;
                  end if;
                  Size := Size + 1;
               else
                  raise Error;
               end if;
            end if;
         end if;
      end loop;

      if Remaining /= 0 then
         raise Error;
      end if;

      return Size;
   end Length_In_UTF16;

   -- ------------------------------------------------------------------------

   function Fits_In_16_Bits (Code : Word24_Type) return Boolean is
   begin
      return (Code and 2#11111111_00000000_00000000#) = 0;
   end Fits_In_16_Bits;

   -- ------------------------------------------------------------------------

   procedure Encode
     (Destination          : API_String_Access;
      Destination_Position : in out Data_Index_Type;
      Last_Part            : Boolean := True;
      Source               : String)
   is
      Destination_Index     : Data_Index_Type;
      Destination_Character : API_Character_Type;
      Remaining             : Natural;
      Byte                  : Byte_Type;
      Code                  : Word24_Type;
   begin

      Destination_Index := Destination_Position;

      if Destination_Index = Data_Index_Type'Last then
         raise Error;
      end if;

      if Source'Length = 0 then
         Destination (Destination_Index) := Null_Terminating_Character;
         Destination_Position := Data_Index_Type'Succ (Destination_Index);
         return;
      end if;

      Remaining := 0;
      Code      := 0;

      for Index in Source'First .. Source'Last loop
         Byte := Character'Pos (Source (Index));
         if Remaining > 0 then

            if (Byte and Hight_Bits_Mask) /= Hight_Bits_Value then
               raise Error;
            end if;

            Byte := Byte and Low_Bits_Mask;
            Code := Code * (2 ** 6);
            Code := Code or Word24_Type (Byte);
            Remaining := Remaining - 1;

            if Remaining = 0 then
               if Fits_In_16_Bits (Code) then
                  Destination_Character := API_Character_Type'Val (Code);
                  Destination (Destination_Index) := Destination_Character;
                  Destination_Index               := Destination_Index + 1;
                  Code := 0;
               else
                  raise Error;
               end if;
            end if;
         elsif (Byte and Hight_Bit_Mask) = 0 then

            Destination_Character := API_Character_Type'Val (Byte);

            Destination (Destination_Index) := Destination_Character;
            Destination_Index               := Destination_Index + 1;
         else
            Byte := Byte * 2;
            while (Byte and Hight_Bit_Mask) /= 0 loop
               Remaining := Remaining + 1;
               Byte      := Byte * 2;
            end loop;

            if Byte = 0 then
               raise Error;
            end if;

            if Remaining > 4 then
               raise Error;
            elsif Remaining = 0 then
               raise Error;
            end if;

            Byte := Byte / (2 ** (Remaining + 1));
            Code := Word24_Type'Val (Byte);
         end if;
      end loop;

      if Remaining /= 0 then
         raise Error;
      end if;

      if Last_Part then
         Destination (Destination_Index) := Null_Terminating_Character;

         if Destination_Index < Data_Index_Type'Last then
            Destination_Index := Data_Index_Type'Succ (Destination_Index);
         else
            Destination_Index := Data_Index_Type'Last;
         end if;
      end if;

      Destination_Position := Destination_Index;

   end Encode;

   -- ------------------------------------------------------------------------

   Empty_Source_String : constant String := "";

   -- ------------------------------------------------------------------------

   function Create
     (Source        : String;
      Null_If_Empty : Boolean := False)
      return API_String_Type
   is
      Data_Length    : Storage_Units_Count_Type;
      Write_Position : Data_Index_Type;
   begin
      if Source = Empty_Source_String then
         if Null_If_Empty then
            return Null_String;
         end if;
      end if;

      return Result : API_String_Type do
         Data_Length        := Length_In_UTF16 (Source);
         Result.String_Data := Free_Static_Allocation_For (Data_Length);
         Result.Data_Length := Data_Length;
         Write_Position     := Result.String_Data'First;
         Encode
           (Destination          => Result.String_Data,
            Destination_Position => Write_Position,
            Source               => Source);
      end return;
   end Create;

   -- ------------------------------------------------------------------------

   Environment_Name_Value_Separator     : constant String := "=";
   Environment_Block_Terminating_String : constant String := "";

   -- ------------------------------------------------------------------------

   function Create
     (Source        : Environment_Variables.Instance_Type;
      Null_If_Empty : Boolean := False)
      return          Environment_Block_Type
   is
   begin
      if Environment_Variables.Is_Empty (Source) then
         if Null_If_Empty then
            return Null_Environment_Block;
         end if;
      end if;

      declare
         function Environment_Block_Length return Storage_Units_Count_Type
         is
         begin
            return Result : Storage_Units_Count_Type := 0 do
               declare
                  Environment_Item : Environment_Variables.Iterator_Type :=
                    (Environment_Variables.First (Source));
               begin
                  while Environment_Variables.Not_Null (Environment_Item) loop
                     declare
                        Name_Length : constant Storage_Units_Count_Type :=
                          (Length_In_UTF16
                             (Environment_Variables.Name
                                (Environment_Item)));

                        Separator_Length : constant Storage_Units_Count_Type :=
                          (Length_In_UTF16
                             (Environment_Name_Value_Separator));

                        Value_Length : constant Storage_Units_Count_Type :=
                          (Length_In_UTF16
                             (Environment_Variables.Value
                                (Environment_Item)));

                        Chunk_Length : Storage_Units_Count_Type := 0;
                     begin
                        Chunk_Length := Name_Length;

                        if
                          (Storage_Units_Count_Type'Last - Chunk_Length) <
                          (Separator_Length)
                        then
                           raise Error;
                        end if;

                        Chunk_Length := Chunk_Length + Separator_Length;

                        if
                          (Storage_Units_Count_Type'Last - Chunk_Length) <
                          (Value_Length)
                        then
                           raise Error;
                        end if;

                        Chunk_Length := Chunk_Length + Value_Length;

                        if ((Result) >=
                              (Storage_Units_Count_Type'Last - Chunk_Length))
                        then
                           raise Error;
                        end if;

                        Result := Result + Chunk_Length;
                        Result := Storage_Units_Count_Type'Succ (Result);
                     end;
                     Environment_Variables.Move_To_Next (Environment_Item);
                  end loop;

                  if Result = Storage_Units_Count_Type'Last then
                     raise Error;
                  end if;

               end;
            end return;
         end;

         Data_Length : constant Storage_Units_Count_Type :=
           (Environment_Block_Length);

         Write_Position : Data_Index_Type;

      begin

         return Result : Environment_Block_Type do
            Result.String_Data := Free_Static_Allocation_For (Data_Length);
            Result.Data_Length := Data_Length;
            Write_Position := Result.String_Data'First;

            declare
               Environment_Item : Environment_Variables.Iterator_Type :=
                 (Environment_Variables.First (Source));
            begin
               while Environment_Variables.Not_Null (Environment_Item) loop
                  Encode
                    (Destination          => Result.String_Data,
                     Destination_Position => Write_Position,
                     Last_Part            => False,
                     Source               =>
                       (Environment_Variables.Name
                          (Environment_Item)));
                  Encode
                    (Destination          => Result.String_Data,
                     Destination_Position => Write_Position,
                     Last_Part            => False,
                     Source               =>
                       (Environment_Name_Value_Separator));
                  Encode
                    (Destination          => Result.String_Data,
                     Destination_Position => Write_Position,
                     Last_Part            => True,
                     Source               =>
                       (Environment_Variables.Value
                          (Environment_Item)));
                  Environment_Variables.Move_To_Next (Environment_Item);
               end loop;
               Encode
                 (Destination          => Result.String_Data,
                  Destination_Position => Write_Position,
                  Last_Part            => True,
                  Source               =>
                    (Environment_Block_Terminating_String));
            end;
         end return;
      end;
   end Create;

   -- ------------------------------------------------------------------------

   function Null_String return API_String_Type is
   begin
      return Result : API_String_Type do
         Result.String_Data := null;
         Result.Data_Length := 0;
      end return;
   end Null_String;

   -- ------------------------------------------------------------------------

   function Null_String return API_String_Access is
   begin
      return null;
   end Null_String;

   -- ------------------------------------------------------------------------

   function Null_Environment_Block return Environment_Block_Type is
   begin
      return Result : Environment_Block_Type do
         Result.String_Data := null;
         Result.Data_Length := 0;
      end return;
   end Null_Environment_Block;

   -- ------------------------------------------------------------------------

   function Null_Environment_Block return Environment_Block_Access is
   begin
      return null;
   end Null_Environment_Block;

   -- ------------------------------------------------------------------------

   function Access_To
     (API_String : API_String_Type)
      return       API_String_Access
   is
   begin
      return API_String.String_Data;
   end Access_To;

   -- ------------------------------------------------------------------------

   function Access_To_Or_Null
     (API_String     : API_String_Type;
      Null_Condition : Boolean)
      return       API_String_Access
   is
   begin
      if not Null_Condition then
         return Access_To (API_String);
      else
         return null;
      end if;
   end;

   -- ------------------------------------------------------------------------

   function Access_To
     (Environment_Block : Environment_Block_Type)
      return              Environment_Block_Access
   is
   begin
      return Environment_Block_Access (Environment_Block.String_Data);
   end Access_To;

   -- ------------------------------------------------------------------------

   function Access_To_Or_Null
     (Environment_Block : Environment_Block_Type;
      Null_Condition    : Boolean)
      return              Environment_Block_Access
   is
   begin
      if not Null_Condition then
         return Access_To (Environment_Block);
      else
         return null;
      end if;
   end;

   -- ------------------------------------------------------------------------

   function Data_Length
     (API_String : API_String_Type)
      return       Data_Length_Type
   is
   begin
      if API_String.String_Data = null then
         raise Error;
      end if;

      return API_String.Data_Length;
   end Data_Length;

   -- ------------------------------------------------------------------------

begin
   Static_Allocations (1).Data := new
     String_Data_Type'(1 .. 25 => API_Character_Type'Val (0));
   Static_Allocations (2).Data := new
     String_Data_Type'(1 .. 25 => API_Character_Type'Val (0));
   Static_Allocations (3).Data := new
     String_Data_Type'(1 .. 90 => API_Character_Type'Val (0));
   Static_Allocations (4).Data := new
     String_Data_Type'(1 .. 100 => API_Character_Type'Val (0));
   Static_Allocations (5).Data := new
     String_Data_Type'(1 .. 150 => API_Character_Type'Val (0));
   Static_Allocations (6).Data := new
     String_Data_Type'(1 .. 300 => API_Character_Type'Val (0));
   Static_Allocations (7).Data := new
     String_Data_Type'(1 .. 400 => API_Character_Type'Val (0));
   Static_Allocations (8).Data := new
     String_Data_Type'(1 .. 2048 => API_Character_Type'Val (0));
   Static_Allocations (9).Data := new
     String_Data_Type'(1 .. 8192 => API_Character_Type'Val (0));
end Windows.Strings;
